# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import numpy as np

from ._kinsol_numpy import solve as _solve


def solve(f_cb, j_cb, x0, fnormtol=1e-6, scsteptol=1e-12, x_scale=None,
          f_scale=None, constraints=None, lband=-1, uband=-1, mxiter=200):
    """
    Solves a system of (non-linear) equations.

    Parameters
    ----------
    f_cb: callable
        Function with signature f(x, fout) which modifies fout *inplace*.
    j_cb: callable
        Function with signature j(x, jmat_out, fx) which modifies
        jmat_out *inplace*. fx contains current fout of f_cb(x).
    x0: array_like
        initial guess
    \*\*kwargs: dict

    Returns
    -------
    dict:
        'x': solution vector
        'success': bool
        'nfev': int
        'njev': int
    """
    x0 = np.asarray(x0, dtype=np.float64)
    if f_scale is None:
        f_scale = np.ones(x0.size)
    if x_scale is None:
        x_scale = np.ones(x0.size)
    if constraints is None:
        constraints = np.zeros(x0.size)
    return _solve(f_cb, j_cb, x0, fnormtol, scsteptol, x_scale, f_scale,
                  constraints, lband, uband, mxiter)



# -*- coding: utf-8 -*-
"""
pykinsol provides a Python interface to KINSOL from the SUNDIALS package suite.
"""
from __future__ import division, absolute_import

from ._release import __version__
from .core import solve



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np


def _get_jmat_out(ny, lband=None, uband=None):
    if lband is not None:
        return np.empty((1+lband+uband, ny))
    else:
        return np.empty((ny, ny))


def _check_callable(f, j, x0, y0, lband=None, uband=None):
    ny = len(y0)
    _fout = np.empty(ny)
    _ret = f(x0, y0, _fout)
    if _ret is not None:
        raise ValueError("f() must return None")

    if j is None:
        return  # Not all methods require a jacobian

    _jmat_out = _get_jmat_out(ny, lband, uband)
    _dfdx_out = np.empty(ny)
    _ret = j(x0, y0, _jmat_out, _dfdx_out)
    if _ret is not None:
        raise ValueError("j() must return None")


def _check_indexing(f, j, x0, y0, lband=None, uband=None):
    ny = len(y0)
    _fout_short = np.empty(ny-1)
    try:
        f(x0, y0, _fout_short)
    except (IndexError, ValueError):
        pass
    else:
        raise ValueError("All elements in fout not assigned in f()")

    if j is None:
        return  # Not all methods require a jacobian

    _dfdx_out = np.empty(ny)
    _jmat_out_short = np.empty((ny, ny-1))
    try:
        j(x0, y0, _jmat_out_short, _dfdx_out)
    except (IndexError, ValueError):
        pass
    else:
        raise ValueError("All elements in Jout not assigned in j()")

    _jmat_out = _get_jmat_out(ny, lband, uband)
    _dfdx_out_short = np.empty(ny-1)
    try:
        j(x0, y0, _jmat_out, _dfdx_out_short)
    except (IndexError, ValueError):
        pass
    else:
        raise ValueError("All elements in dfdx_out not assigned in j()")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Tested with Sundials 2.6.2

import os
import sys
from setuptools import setup, Extension


pkg_name = 'pykinsol'

# Cythonize .pyx file if it exists (not in source distribution)
ext_modules = []

LLAPACK = os.environ.get('LLAPACK', 'lapack')

if len(sys.argv) > 1 and '--help' not in sys.argv[1:] and sys.argv[1] not in (
        '--help-commands', 'egg_info', 'clean', '--version'):
    import numpy as np
    USE_CYTHON = os.path.exists('pykinsol/_kinsol_numpy.pyx')  # not in sdist
    ext = '.pyx' if USE_CYTHON else '.cpp'
    ext_modules = [
        Extension('pykinsol._kinsol_numpy',
                  ['pykinsol/_kinsol_numpy'+ext],
                  language='c++', extra_compile_args=['-std=c++11'],
                  libraries=['sundials_kinsol', LLAPACK,
                             'sundials_nvecserial'],
                  include_dirs=[np.get_include(), './include'])
    ]
    if USE_CYTHON:
        from Cython.Build import cythonize
        ext_modules = cythonize(ext_modules, include_path=['./include'])

PYKINSOL_RELEASE_VERSION = os.environ.get('PYKINSOL_RELEASE_VERSION', '')

# http://conda.pydata.org/docs/build.html#environment-variables-set-during-the-build-process
CONDA_BUILD = os.environ.get('CONDA_BUILD', '0') == '1'
if CONDA_BUILD:
    try:
        PYKINSOL_RELEASE_VERSION = 'v' + open(
            '__conda_version__.txt', 'rt').readline().rstrip()
    except IOError:
        pass

release_py_path = os.path.join(pkg_name, '_release.py')

if (len(PYKINSOL_RELEASE_VERSION) > 1 and
   PYKINSOL_RELEASE_VERSION[0] == 'v'):
    TAGGED_RELEASE = True
    __version__ = PYKINSOL_RELEASE_VERSION[1:]
else:
    TAGGED_RELEASE = False
    # read __version__ attribute from _release.py:
    exec(open(release_py_path).read())

classifiers = [
    "Development Status :: 3 - Alpha",
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Mathematics',
]

tests = [
    'pykinsol.tests',
]

descr = 'Python binding for kinsol from the sundials library.'
with open(pkg_name + '/__init__.py') as f:
    long_description = f.read().split('"""')[1]

setup_kwargs = dict(
    name=pkg_name,
    version=__version__,
    description=descr,
    long_description=long_description,
    classifiers=classifiers,
    author='Björn Dahlgren',
    author_email='bjodah@DELETEMEgmail.com',
    url='https://github.com/bjodah/' + pkg_name,
    license='BSD',
    packages=[pkg_name] + tests,
    ext_modules=ext_modules,
    install_requires=['numpy']
)

if __name__ == '__main__':
    import shutil
    try:
        if TAGGED_RELEASE:
            # Same commit should generate different sdist
            # depending on tagged version (set PYKINSOL_RELEASE_VERSION)
            # this will ensure source distributions contain the correct version
            shutil.move(release_py_path, release_py_path+'__temp__')
            open(release_py_path, 'wt').write(
                "__version__ = '{}'\n".format(__version__))
        setup(**setup_kwargs)
    finally:
        if TAGGED_RELEASE:
            shutil.move(release_py_path+'__temp__', release_py_path)



# -*- coding: utf-8 -*-

from pykinsol import solve


def f_cb(x, fout):
    # expression from example in SciPy documentation:
    # docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html
    fout[0] = x[0] + (x[0] - x[1])**3/2 - 1
    fout[1] = x[1] + (x[1] - x[0])**3/2


def j_cb(x, Jout, fx):
    Jout[0, 0] = 1 + 3/2 * (x[0] - x[1])**2
    Jout[0, 1] = -3/2 * (x[0] - x[1])**2
    Jout[1, 0] = -3/2 * (x[1] - x[0])**2
    Jout[1, 1] = 1 + 3/2 * (x[1] - x[0])**2


# def test_solve__failure():
#     result = solve(f_cb, j_cb, [1, 0])
#     assert not result['success']  # newton method fails for this


def test_solve():
    result = solve(f_cb, j_cb, [0, 0])
    print(result)
    assert result['success']
    assert result['status'] == 0
    assert result['nit'] > 1
    assert result['nfev'] > 1
    assert result['njev'] > 0
    assert result['time_cpu'] > 1e-15
    assert abs(result['x'][0] - 0.8411639) < 2e-7
    assert abs(result['x'][1] - 0.1588361) < 2e-7
